from typing import Optional

from tikkie._session import ApiSession, get_session
from tikkie.v2.models import CreateSandboxApplicationResponse


def create_application(
    app_token: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[ApiSession] = None,
) -> CreateSandboxApplicationResponse:
    """
    Creates an app in the sandbox for you to use with this API. This
    operation does not work in the production environment.

    https://developer.abnamro.com/api-products/tikkie/reference-documentation#tag/Sandbox-app

    :param app_token: The App Token generated in the Tikkie Business Portal.
    :param api_key: The API Key obtained from the ABN developer portal.
    :param session: An optional API session class. If not passed the global API session will be used
    """
    s = session if session else get_session()
    response = s.post(
        "sandboxapps",
        app_token=app_token,
        api_key=api_key,
    )
    return CreateSandboxApplicationResponse.parse_raw(response.content)



import json
from datetime import date, datetime, time, timezone
from decimal import Decimal
from typing import Any, Dict, Union
from uuid import UUID

from pydantic import BaseModel

JsonDict = Dict[str, Any]
Floatable = Union[Decimal, int, float]


def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() for x in components[1:])


class CamelCaseModel(BaseModel):
    class Config:
        alias_generator = to_camel_case
        allow_population_by_field_name = True


def is_aware(value: Union[datetime, time]) -> bool:
    """
    Determine if a given datetime.datetime is aware.

    The concept is defined in Python's docs:
    https://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """
    return value.utcoffset() is not None


def format_date(d: date) -> str:
    """
    Converts a date into a ISO 8601 (YYYY-MM-DD) format
    """
    return d.isoformat()


def format_datetime(d: datetime) -> str:
    """
    Converts a timezone aware datetime object into a UTC ISO 8601 (Zulu time) string
    """
    if not is_aware(d):
        raise Exception("Naive datetime not supported")
    return d.astimezone(timezone.utc).isoformat(timespec="milliseconds").replace("+00:00", "Z")


class JSONEncoder(json.JSONEncoder):
    """
    Shameless copied from the Django JSON Encoder
    """

    def default(self, o: Any) -> Any:
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime):
            r = o.isoformat()
            if o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith("+00:00"):
                r = r[:-6] + "Z"
            return r
        elif isinstance(o, date):
            return o.isoformat()
        elif isinstance(o, time):
            if is_aware(o):
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r
        elif isinstance(o, (Decimal, UUID)):
            return str(o)
        else:
            return super().default(o)



from typing import Optional

from tikkie._session import ApiSession, get_session
from tikkie.v2.models import Refund


def create(
    *,
    payment_request_token: str,
    payment_token: str,
    description: str,
    amount_in_cents: int,
    reference_id: Optional[str] = None,
    app_token: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[ApiSession] = None,
) -> Refund:
    """
    Creates a refund for a specific payment

    The maximum amount which can be refunded for a payment is the
    amount paid plus €25.00. The application must have payment request
    permission and the refund permission to complete this request.

    https://developer.abnamro.com/api-products/tikkie/reference-documentation#tag/Refund

    :param payment_request_token: Token identifying the payment request.
    :param payment_token: Token identifying one payment of a payment request
    :param description: Description of the payment request, which the payer or payers see.
    :param amount_in_cents: Amount in cents of the payment request (euros). If this value is not specified, an open payment request is created.
    :param reference_id: ID for the reference of the API consumer.
    :param app_token: The App Token generated in the Tikkie Business Portal.
    :param api_key: The API Key obtained from the ABN developer portal.
    :param session: An optional API session class. If not passed the global API session will be used
    """
    s = session if session else get_session()
    payload = {
        "description": description,
        "amountInCents": amount_in_cents,
    }
    if reference_id is not None:
        payload["referenceId"] = reference_id

    response = s.post(
        f"paymentrequests/{payment_request_token}/payments/{payment_token}/refunds",
        json=payload,
        app_token=app_token,
        api_key=api_key,
    )
    return Refund.parse_raw(response.content)


def get(
    *,
    payment_request_token: str,
    payment_token: str,
    refund_token: str,
    app_token: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[ApiSession] = None,
) -> Refund:
    """
    Retrieves details of a specific refund based on a token value.
    The application must have payment request permission to complete
    this request.

    https://developer.abnamro.com/api-products/tikkie/reference-documentation#operation/getRefund

    :param payment_request_token: Token identifying the payment request.
    :param payment_token: Token identifying one payment of a payment request
    :param refund_token: Token identifying one refund of a payment.
    :param app_token: The App Token generated in the Tikkie Business Portal.
    :param api_key: The API Key obtained from the ABN developer portal.
    :param session: An optional API session class. If not passed the global API session will be used
    """
    s = session if session else get_session()
    response = s.get(
        f"paymentrequests/{payment_request_token}/payments/{payment_token}/refunds/{refund_token}",
        app_token=app_token,
        api_key=api_key,
    )
    return Refund.parse_raw(response.content)



from typing import Optional

from tikkie._session import ApiSession, get_session
from tikkie.v2.models import SubscribeResponse


def subscribe(
    *,
    url: str,
    app_token: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[ApiSession] = None,
) -> SubscribeResponse:
    """
    Subscribes to transactions related notifications. One transaction
    subscription can be active at a time. When this request is
    repeated, the existing transaction subscription is overwritten.
    The application must have transactions permission to complete
    this request.

    https://developer.abnamro.com/api-products/tikkie/reference-documentation#tag/Transactions-notification

    :param url: URL where notifications must be sent using a webhook or callback.
    :param app_token: The App Token generated in the Tikkie Business Portal.
    :param api_key: The API Key obtained from the ABN developer portal.
    :param session: An optional API session class. If not passed the global API session will be used
    """
    s = session if session else get_session()
    payload = {"url": url}

    response = s.post(
        "transactionssubscription",
        json=payload,
        app_token=app_token,
        api_key=api_key,
    )
    return SubscribeResponse.parse_raw(response.content)


def delete(
    app_token: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[ApiSession] = None,
) -> None:
    """
    Deletes a subscription. The application must have transactions
    permission to complete this request.

    https://developer.abnamro.com/api-products/tikkie/reference-documentation#operation/subscribeTransactionsNotifications

    :param app_token: The App Token generated in the Tikkie Business Portal.
    :param api_key: The API Key obtained from the ABN developer portal.
    :param session: An optional API session class. If not passed the global API session will be used
    """
    s = session if session else get_session()
    s.delete(
        "transactionssubscription",
        app_token=app_token,
        api_key=api_key,
    )



from datetime import datetime
from typing import Optional

from tikkie._session import ApiSession, get_session
from tikkie._util import JsonDict
from tikkie.v2.models import GetAllPaymentsResponse, Payment


def get(
    payment_request_token: str,
    payment_token: str,
    app_token: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[ApiSession] = None,
) -> Payment:
    """
    Retrieves details of a specific payment made against a specific payment request based on the value of the tokens.
    The application must have payment request permission to complete this request.

    https://developer.abnamro.com/api-products/tikkie/reference-documentation#operation/getPayment

    :param payment_request_token: Token identifying the payment request.
    :param payment_token: Token identifying one payment of a payment request
    :param app_token: The App Token generated in the Tikkie Business Portal.
    :param api_key: The API Key obtained from the ABN developer portal.
    :param session: An optional API session class. If not passed the global API session will be used
    """
    s = session if session else get_session()
    response = s.get(
        f"paymentrequests/{payment_request_token}/payments/{payment_token}",
        app_token=app_token,
        api_key=api_key,
    )
    return Payment.parse_raw(response.content)


def get_all(
    *,
    payment_request_token: str,
    page_size: int,
    page_number: int,
    from_date_time: Optional[datetime] = None,
    to_date_time: Optional[datetime] = None,
    app_token: Optional[str] = None,
    api_key: Optional[str] = None,
    session: Optional[ApiSession] = None,
) -> GetAllPaymentsResponse:
    """
    Retrieves payments made against a specific payment request based on the parameters provided.
    The application must have payment request permission to complete this request.

    https://developer.abnamro.com/api-products/tikkie/reference-documentation#tag/Payment

    :param payment_request_token: Token identifying the payment request.
    :param page_size: Number of the page to be returned.
    :param page_number: Number of items on a page.
    :param from_date_time:The time you start searching for items. Refers to the creation date for payment requests and iDeal QRs, and refers to the bundle date for transaction bundles.
    :param to_date_time: The time you stop searching for items. Refers to the creation date for payment requests and iDeal QRs, and refers to the bundle date for transaction bundles.
    :param app_token: The App Token generated in the Tikkie Business Portal.
    :param api_key: The API Key obtained from the ABN developer portal.
    :param session: An optional API session class. If not passed the global API session will be used
    """
    s = session if session else get_session()
    params: JsonDict = {
        "pageSize": page_size,
        "pageNumber": page_number,
    }
    if from_date_time is not None:
        params["fromDateTime"] = from_date_time
    if to_date_time is not None:
        params["toDateTime"] = to_date_time

    response = s.get(
        f"paymentrequests/{payment_request_token}/payments",
        params=params,
        app_token=app_token,
        api_key=api_key,
    )
    return GetAllPaymentsResponse.parse_raw(response.content)



import os
from functools import lru_cache
from typing import Any, Optional

import requests
from requests import Response

from tikkie._version import VERSION
from tikkie.errors import BaseError

API_KEY_ENV = "TIKKIE_API_KEY"
APP_TOKEN_ENV = "TIKKIE_APP_TOKEN"
API_URL_ENV = "TIKKIE_API_URL"


class ApiSession(requests.Session):
    def __init__(self, *, app_token: str, api_key: str, api_url: str):
        super().__init__()
        self.api_url = api_url
        self.app_token = app_token
        self.api_key = api_key
        self.headers["User-Agent"] = f"Python Tikkie API/{VERSION}"

    def prepare_request(self, request: Any) -> Any:
        request.url = "/".join(s.strip("/") for s in [self.api_url, "v2/tikkie", request.url])
        return super().prepare_request(request)

    def request(
        self,
        *args: Any,
        params: Any = None,
        app_token: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs: Any,
    ) -> Response:
        dict_params = params or {}
        headers = kwargs.pop("headers", {})

        headers["X-App-Token"] = app_token or self.app_token
        headers["API-Key"] = api_key or self.api_key

        # Manually convert the json objects as we want to use a custom encoder
        # if "json" in kwargs:
        #     payload = kwargs.pop("json")
        #     kwargs["data"] = json.dumps(payload, cls=JSONEncoder)
        #     headers["Content-Type"] = "application/json"

        r = super().request(*args, params=dict_params, headers=headers, **kwargs)  # type: ignore
        if not r.ok:
            raise BaseError.from_response(r)
        return r


# The global session object
_session: Optional[ApiSession] = None


def _read_from_os(var: Optional[str], env: str) -> Optional[str]:
    if var is None:
        var = os.environ.get(env)
    return var


@lru_cache()
def get_session() -> ApiSession:
    global _session
    if _session is None:
        configure()
    return _session  # type: ignore


def configure(
    *, app_token: Optional[str] = None, api_key: Optional[str] = None, api_url: Optional[str] = None
) -> None:
    global _session

    app_token = _read_from_os(app_token, APP_TOKEN_ENV)
    api_key = _read_from_os(api_key, API_KEY_ENV)
    api_url = _read_from_os(api_url, API_URL_ENV)

    if not api_url:
        raise Exception("The API Url is not set")

    get_session.cache_clear()

    _session = ApiSession(app_token=app_token, api_url=api_url, api_key=api_key)
